/*******************************************************************************

   a simple traffic simulation on a 16 x 16 Neopixel grid

   by noiasca

   2021-05-15
   2020-02-13

 *******************************************************************************/

// NEOPIXEL BEST PRACTICES for most reliable operation:
// - AVOID connecting NeoPixels on a LIVE CIRCUIT. If you must, ALWAYS
//   connect GROUND (-) first, then +, then data.

#include <Adafruit_NeoPixel.h>
#define LED_COUNT 256 // How many NeoPixels are attached to the Arduino? (max is 256 for this sketch!)
#define LED_PIN    12 // Which pin on the Arduino is connected to the NeoPixels?
Adafruit_NeoPixel strip(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ800); // Declare our NeoPixel strip object

const byte north = 0;
const byte east = 1;
const byte south = 2;
const byte west = 3;

enum States {RUIN, PARK, HOUSE, FIRE};

#define trafficRed    0xFF0000                                // a color for red traffic signs
#define trafficYellow 0xFFDD00                                // a color for yellow traffic signs
#define trafficGreen  0                                       // a color for green traffic signs - kept black - as currently each color stops cars
#define houseColor    0x404040
#define ruinColor     0xFF4000 // 0xB88604          // 0xAA5500
#define burningColor  0x400000
#define parkColor     0x00AA00

const byte parkPin = A0;
const byte ruinPin = A1;
//A2 currently unused pin
const byte firePin = A3;

struct coordinateType {
  byte x;
  byte y;
};

const byte houses[][2] PROGMEM = {
  // we need two parameters for each house:
  // pos: the tile/pixel/position where the house is located on
  // street: the address of the street which belongs to the house
  // pos, street
  {34, 29},
  {35, 28},
  {36, 27},
  {37, 26},
  {42, 41},
  {43, 20},
  {44, 19},
  {45, 18},
  {50, 49},
  //{51, 49}, // this would enable a house in the center of the block, but it will still need an accessible street (address)
  //{52, 54},
  {53, 54},
  {58, 57},
  //{59, 57},
  //{60, 62},
  {61, 62},
  {66, 65},
  //{67, 65},
  //{68, 70},
  {69, 70},
  {74, 73},
  //{75, 73},
  //{76, 78},
  {77, 78},
  {82, 81},
  {83, 108},
  {84, 107},
  {85, 106},
  {90, 89},
  {91, 100},
  {92, 99},
  {93, 98},
  {162, 161},
  {163, 156},
  {164, 155},
  {165, 154},
  {170, 169},
  {172, 147},
  {171, 148},
  {173, 146},
  {178, 177},
  {181, 182},
  {186, 185},
  {189, 190},
  {194, 193},
  {197, 198},
  {202, 201},
  {205, 206},
  {210, 237},
  {211, 236},
  {212, 235},
  {213, 234},
  {218, 217},
  {219, 228},
  {220, 227},
  {221, 226}
};
const byte noOfHouses = sizeof(houses) / sizeof(houses[0]);
const byte noOfDesasters = 8;        // maximum number of desasters the city can handle. Missing tbd: bring into class City

// The streets need waypoint definition where each street/pixelcan go to:
const byte street [][4] PROGMEM = {
  {0, 0, 0, 1},    // 00  // pixel 0 will not be reached but we still keep it in the array to start with a simple 0
  {0, 0, 0, 2},    // 01  // the order of targets is noted clockwise: North, East, South, West. So the target 2 is in the West.
  {0, 0, 0, 3},    // 02
  {0, 0, 0, 4},    // 03
  {0, 0, 0, 5},    // 04
  {0, 0, 0, 6},    // 05
  {0, 0, 0, 7},    // 06
  {0, 0, 23, 0},   // 07  // here we force each car to head South to get more traffic to the city center
  {0, 0, 0, 9},    // 08
  {0, 0, 0, 10},   // 09
  {0, 0, 0, 11},   // 10
  {0, 0, 0, 12},   // 11
  {0, 0, 0, 13},   // 12
  {0, 0, 0, 14},   // 13
  {0, 0, 0, 15},   // 14
  {0, 0, 16, 0},   // 15
  {0, 0, 47, 0},   // 16
  {0, 18, 0, 0},   // 17
  {0, 19, 0, 0},   // 18
  {0, 20, 0, 0},   // 19
  {0, 21, 0, 0},   // 20
  {0, 22, 0, 0},   // 21
  {0, 23, 41, 0},  // 22
  {0, 24, 0, 0},   // 23
  {8, 25, 0, 0},   // 24
  {0, 26, 0, 0},   // 25
  {0, 27, 0, 0},   // 26
  {0, 28, 0, 0},   // 27
  {0, 29, 0, 0},   // 28
  {0, 30, 0, 0},   // 29
  {0, 0, 33, 0},   // 30
  {1, 0, 0, 0},    // 31 edge case: we leave pixel 0 empty (north) and jump to pixel 1
  {31, 0, 0, 0},   // 32
  {0, 0, 62, 0},   // 33
  {0, 0, 0, 0},    // 34 house: this field is used as house address, therefore it can't be reached and it doesn't need any targets
  {0, 0, 0, 0},    // 35
  {0, 0, 0, 0},    // 36
  {0, 0, 0, 0},    // 37
  {25, 0, 0, 0},   // 38
  {24, 0, 0, 0},   // 39
  {0, 0, 55, 54},  // 40
  {0, 55, 54, 0},  // 41
  {0, 0, 0, 0},    // 42
  {0, 0, 0, 0},    // 43
  {0, 0, 0, 0},    // 44
  {0, 0, 0, 0},    // 45
  {17, 0, 0, 0},   // 46
  {0, 0, 48, 0},   // 47
  {0, 0, 79, 0},   // 48
  {46, 0, 0, 0},   // 49
  {0, 0, 0, 0},    // 50
  {0, 0, 0, 0},    // 51
  {0, 0, 0, 0},    // 52
  {0, 0, 0, 0},    // 53
  {0, 0, 73, 0},   // 54
  {0, 0, 72, 0},   // 55
  {39, 0, 0, 0},   // 56
  {38, 0, 0, 0},   // 57
  {0, 0, 0, 0},    // 58
  {0, 0, 0, 0},    // 59
  {0, 0, 0, 0},    // 60
  {0, 0, 0, 0},    // 61
  {0, 0, 65, 0},   // 62
  {32, 0, 0, 0},   // 63
  {63, 0, 0, 0},   // 64
  {0, 0, 94, 0},   // 65
  {0, 0, 0, 0},    // 66
  {0, 0, 0, 0},    // 67
  {0, 0, 0, 0},    // 68
  {0, 0, 0, 0},    // 69
  {57, 0, 0, 0},   // 70
  {56, 0, 0, 0},   // 71
  {0, 0, 87, 0},   // 72
  {0, 0, 86, 0},   // 73
  {0, 0, 0, 0},    // 74
  {0, 0, 0, 0},    // 75
  {0, 0, 0, 0},    // 76
  {0, 0, 0, 0},    // 77
  {49, 0, 0, 0},   // 78
  {0, 0, 80, 0},   // 79
  {0, 0, 111, 0},  // 80
  {78, 0, 0, 0},   // 81
  {0, 0, 0, 0},    // 82
  {0, 0, 0, 0},    // 83
  {0, 0, 0, 0},    // 84
  {0, 0, 0, 0},    // 85
  {0, 0, 105, 0},  // 86
  {0, 0, 104, 0},  // 87
  {71, 70, 0, 0},  // 88
  {70, 0, 0, 71},  // 89
  {0, 0, 0, 0},    // 90
  {0, 0, 0, 0},    // 91
  {0, 0, 0, 0},    // 92
  {0, 0, 0, 0},    // 93
  {0, 0, 97, 0},   // 94
  {64, 0, 0, 0},   // 95
  {95, 0, 0, 0},   // 96
  {0, 0, 126, 98}, // 97
  {0, 0, 124, 99}, // 98
  {0, 0, 0, 100},  // 99
  {0, 0, 0, 101},  // 100
  {0, 0, 0, 102},  // 101
  {89, 0, 0, 103}, // 102
  {88, 0, 0, 104}, // 103
  {0, 0, 119, 105},// 104
  {0, 0, 118, 106},// 105
  {0, 0, 116, 107},// 106
  {0, 0, 0, 108},  // 107
  {0, 0, 0, 109},  // 108
  {0, 0, 0, 110},  // 109
  {81, 0, 0, 0},   // 110
  {0, 0, 112, 0},  // 111
  {0, 142, 143, 0},// 112
  {110, 143, 0, 0},// 113
  {0, 0, 0, 113},  // 114
  {0, 0, 0, 114},  // 115
  {0, 0, 0, 115},  // 116
  {107, 0, 0, 116},// 117
  {0, 0, 137, 117},// 118
  {0, 135, 136, 118},    // 119
  {103, 0, 0, 119},      // 120
  {102, 0, 135, 120},    // 121
  {0, 0, 0, 121},  // 122
  {0, 0, 0, 122},  // 123
  {0, 0, 0, 123},  // 124
  {99, 0, 0, 124}, // 125
  {0, 0, 129, 125},// 126
  {96, 0, 0, 0},   // 127
  {127, 0, 0, 126},// 128   // eventuell nur abzweigen zulassen
  {127, 0, 158, 0},// 129
  {0, 0, 0, 129},  // 130
  {0, 0, 0, 130},  // 131
  {0, 0, 0, 131},  // 132
  {0, 155, 0, 132},    // 133
  {121, 0, 0, 133},    // 134
  {120, 0, 119, 134},  // 135
  {120, 151, 0, 135},    // 136
  {0, 150, 0, 136},    // 137  - linksabbieger vs. Nord-Süd fahrer
  {0, 0, 0, 137},  // 138
  {0, 0, 0, 138},  // 139
  {0, 0, 0, 139},  // 140
  {0, 0, 147, 140},// 141
  {113, 0, 0, 141},// 142
  {0, 0, 144, 0},  // 143
  {0, 0, 175, 0},  // 144
  {142, 146, 0, 0},  // 145
  {140, 0, 0 , 147}, // 146
  {0, 0, 0, 148},  // 147
  {0, 0, 0, 149},  // 148
  {0, 0, 0, 150},  // 149
  {0, 169, 0, 151},// 150
  {0, 168, 0, 152},// 151
  {135, 0, 0, 153},// 152
  {134, 0, 0, 154},  // 153
  {132, 0, 0, 155},  // 154
  {131, 0, 0, 156},  // 155  mysterious street
  {0, 0, 0, 157},  // 156
  {0, 0, 0, 158},  // 157
  {0, 0, 0, 161},  // 158
  {128, 0, 0, 0},  // 159
  {159, 0, 0, 0},  // 160
  {0, 0, 190, 0},  // 161
  {0, 0, 0, 0},    // 162
  {0, 0, 0, 0},    // 163
  {0, 0, 0, 0},    // 164
  {0, 0, 0, 0},    // 165
  {153, 0, 0, 154},// 166
  {152, 0, 0, 0},  // 167
  {0, 0, 183, 182},// 168
  {0, 183, 182, 0},// 169
  {0, 0, 0, 0},    // 170
  {0, 0, 0, 0},    // 171
  {0, 0, 0, 0},    // 172
  {0, 0, 0, 0},    // 173
  {145, 0, 0, 0},  // 174
  {0, 0, 176, 0},  // 175
  {0, 0, 207, 0},  // 176
  {174, 0, 0, 0},  // 177
  {0, 0, 0, 0},    // 178
  {0, 0, 0, 0},    // 179
  {0, 0, 0, 0},    // 180
  {0, 0, 0, 0},    // 181
  {0, 0, 201, 0},  // 182
  {0, 0, 200, 0},  // 183
  {167, 0, 0, 0},  // 184
  {166, 0, 0, 0},  // 185
  {0, 0, 0, 0},    // 186
  {0, 0, 0, 0},    // 187
  {0, 0, 0, 0},    // 188
  {0, 0, 0, 0},    // 189
  {0, 0, 193, 0},  // 190
  {160, 0, 0, 0},  // 191
  {191, 0, 0, 0},  // 192
  {0, 0, 222, 0},  // 193
  {0, 0, 0, 0},    // 194
  {0, 0, 0, 0},    // 195
  {0, 0, 0, 0},    // 196
  {0, 0, 0, 0},    // 197
  {185, 0, 0, 0},  // 198
  {184, 0, 0, 0},  // 199
  {0, 0, 215, 0},  // 200
  {0, 0, 214, 0},  // 201
  {0, 0, 0, 0},    // 202
  {0, 0, 0, 0},    // 203
  {0, 0, 0, 0},    // 204
  {0, 0, 0, 0},    // 205
  {177, 0, 0, 0},  // 206
  {0, 0, 208, 0},  // 207
  {0, 0, 239, 0},  // 208
  {206, 0, 0, 0},  // 209
  {0, 0, 0, 0},    // 210
  {0, 0, 0, 0},    // 211
  {0, 0, 0, 0},    // 212
  {0, 0, 0, 0},    // 213
  {0, 0, 233, 0},  // 214
  {0, 0, 232, 0},  // 215
  {199, 0, 0, 198},// 216
  {198, 199, 0, 0},// 217
  {0, 0, 0, 0},    // 218
  {0, 0, 0, 0},    // 219
  {0, 0, 0, 0},    // 220
  {0, 0, 0, 0},    // 221
  {0, 0, 225, 0},  // 222
  {192, 0, 0, 0},  // 223
  {223, 0, 0, 0},  // 224
  {0, 0, 0, 226},  // 225
  {0, 0, 0, 227},  // 226
  {0, 0, 0, 228},  // 227
  {0, 0, 0, 229},  // 228
  {0, 0, 0, 230},  // 229
  {217, 0, 0, 231},// 230
  {216, 0, 0, 232},// 231
  {0, 248, 0, 233},// 232
  {0, 0, 0, 234},  // 233
  {0, 0, 0, 235},  // 234
  {0, 0, 0, 236},  // 235
  {0, 0, 0, 237},  // 236
  {0, 0, 0, 238},  // 237
  {209, 0, 0, 0},  // 238
  {0, 0, 240, 0},  // 239
  {0, 241, 0, 0},  // 240
  {0, 242, 0, 0},  // 241
  {0, 243, 0, 0},  // 242
  {0, 244, 0, 0},  // 243
  {0, 245, 0, 0},  // 244
  {0, 246, 0, 0},  // 245
  {0, 247, 0, 0},  // 246
  {231, 248, 0, 0},// 247
  {0, 249, 0, 0},  // 248
  {0, 250, 0, 0},  // 249
  {0, 251, 0, 0},  // 250
  {0, 252, 0, 0},  // 251
  {0, 253, 0, 0},  // 252
  {0, 254, 0, 0},  // 253
  {0, 255, 0, 0},  // 254
  {224, 0, 0, 0},  // 255
};


void showMap()               // shows 4 times a 2 x 2 block of houses
{
  const byte houses[] {51, 59, 179, 187};  // start position of each 2 x 2 block
  for (auto & i : houses)
  {
    for (byte j = 0; j < 2; j++)
    {
      strip.setPixelColor(i + j, houseColor);
      strip.setPixelColor((i % 16) + 16 * (i / 16 + 1) + j , houseColor);
    }
  }
  // strip.show(); // will be shown anyway with the first call
}


// calculate x and y coordinates from for a given position
// position 0 is 0,0
coordinateType position2coordinate(byte position)
{
  coordinateType result;
  byte row = position / 16;
  result.y = row;
  if (row % 2)
    result.x = 15 - position % 16;     //even rows in my s-Grid
  else
    result.x = position % 16;          //odd rows in my s-Grid
  return result;
}


// calculates distance between two positions
uint16_t distance(byte positionA, byte positionB)
{
  uint16_t distanceSquare;
  coordinateType coordinateA, coordinateB;
  coordinateA = position2coordinate(positionA);
  coordinateB = position2coordinate(positionB);
  int16_t diffX = coordinateA.x - coordinateB.x;
  int16_t diffY = coordinateA.y - coordinateB.y;
  if (diffX < 0) diffX *= -1;
  if (diffY < 0) diffY *= -1;
  distanceSquare = diffX * diffX + diffY * diffY;          // we don' need the absolute value, the square of the hypotenuse is ok.
  return distanceSquare;
}


class Button
{
    const byte buttonPin;
    const byte debounceDelay = 50;    // the debounce time; increase if the output flickers
    bool lastButtonState = HIGH;      // the previous reading from the input pin
    // the following variable is ONE BYTE only as we don't measure absolute running time,
    byte lastDebounceTime = 0;        // the last time the output pin was toggled - we check only ONE byte

  public:
    Button(byte attachTo) : buttonPin(attachTo) {}

    void begin() {
      pinMode(buttonPin, INPUT_PULLUP);
    }

    // returns HIGH if button was clicked since last call - debounced
    bool wasPressed() {
      bool buttonState = LOW;                  // the current reading from the input pin
      byte reading = !digitalRead(buttonPin);                         // if we are using INPUT_PULLUP we are checking invers to LOW Pin

      if (((millis() & 0xFF ) - lastDebounceTime) > debounceDelay)    // If the switch changed, AFTER any pressing or noise
      {
        if (reading != lastButtonState && lastButtonState == LOW)     // If there was a change and and last state was LOW
        {
          buttonState = HIGH;
        }
        lastDebounceTime = millis() & 0xFF;
        lastButtonState = reading;
      }
      return buttonState;
    }
};


class City                                                 // the city object cares about the houses and the disasters
{
  public:
    uint32_t previousMillis = 0;                           // timekeeping
    byte state[noOfHouses];                                // a state for each house
    struct Disaster {
      byte id;                                             // effected house 0 - 47
      uint32_t previous;                                   // timekeeping for this desaster
    } disaster[noOfDesasters];

    void showHouse(byte id)
    {
      uint32_t color = houseColor;
      switch (state[id])
      {
        case RUIN : color = ruinColor; break;
        case PARK : color = parkColor; break;
        case HOUSE : color = houseColor; break;
        case FIRE : color = burningColor; break;
      }
      strip.setPixelColor(pgm_read_byte(&houses[id][0]), color);
    }

    void showAllHouses()
    {
      for (byte i = 0; i < noOfHouses; i++)
        showHouse(i);
    }

    void receivedFireOut(byte street, byte order)
    {
      byte actual = disaster[order].id;
      //Serial.print(F(" order=")); Serial.print(order);
      //Serial.print(F(" actual=")); Serial.print(actual);
      //Serial.print(F(" street=")); Serial.print(order);
      //Serial.print(F(" pgm=")); Serial.print(pgm_read_byte(&houses[actual][1]));
      if (pgm_read_byte(&houses[actual][1]) == street)
      {
        Serial.print (F(" - street fits to order"));
        if (state[actual] == FIRE)
        {
          Serial.print(F(" - in time"));
          state[actual] = HOUSE;
          showHouse(actual);
          disaster[order].id = 0;
        }
        //Serial.println();
      }
    }

    bool addDisaster(byte i)                                    // try to apply a disaster to a house id = i
    {
      bool success = false;
      for (byte j = 0; j < noOfDesasters; j++)
        if (disaster[j].id == 0) {
          disaster[j].id = i;
          disaster[j].previous = millis();
          state[i]++;
          showHouse(i);
          success = true;
          firedepartmentAlert(pgm_read_byte(&houses[i][1]), j);  // alert the firedepartment
          break;
        }
      return success;
    }

    void tick()                                            // time slot to manage the City
    {
      if (millis() - previousMillis > 500)
      {
        previousMillis = millis();
        for (byte i = 0; i < noOfHouses; i++)
        {
          if (random(1000) == 1)                           // there is a seldom chance that addresses change their status randomly
          {
            if (state[i] < HOUSE) {                        // Upgrade an address RUIN --> PARK --> FIRE
              //Serial.print(F("house state ")); Serial.println(pgm_read_byte(&houses[i][0]));
              state[i]++;
              showHouse(i);
            }
            else if (state[i] == HOUSE)
            {
              addDisaster(i); // if empty disaster as available, generate a disaster
            }
          }
          if (state[i] == FIRE)                             // take care if the fire was extinguished in time
          {
            // search the correct disaster by house id
            for (byte j = 0; j < noOfDesasters; j++)
              if (disaster[j].id == i)
              {
                if (millis() - disaster[j].previous > 30000)
                {
                  disaster[j].id = 0;
                  Serial.print(F("House destroyed @")); Serial.println(pgm_read_byte(&houses[i][0]));
                  state[i] = 0;
                  showHouse(i);
                }
                break;
              }
          }
        }
      }
    }
} city;                                                    // we only need one instance: the city object


class Crossing                                             // a street crossing manager cares for up to 8 traffic lights
{
    byte state = 0;                                        // 0 north/south - 1 yellow - 2 east/west - 3 yellow - MISSING: should use an enum instead...
    byte north0, north1, north2, north3;                   // the pixels used for the max 4 north/south traffic signs.
    byte east0, east1, east2, east3;                       // the pixels used for the max 4 east/west traffic signs.
    uint16_t intervalLong = 5000;                          // intervall in ms for red/green phases
    uint16_t intervalShort = 1500;                         // shorter intervall for "all yellow"
    uint16_t interval = intervalLong;                      // the current interval
    uint32_t previousMillis = 0 - intervalLong;            // timekeeping for intervals

  public:
    Crossing(byte north0, byte north1, byte north2, byte north3, byte east0, byte east1, byte east2, byte east3):
      north0(north0),
      north1(north1),
      north2(north2),
      north3(north3),
      east0(east0),
      east1(east1),
      east2(east2),
      east3(east3)
    {}

    Crossing(byte north0, byte north1, byte east0, byte east1):
      north0(north0),
      north1(north1),
      east0(east0),
      east1(east1)
    {}

    void setState(byte newState) {
      state = newState;
      //strip.show();
      //state change doesn't update the pixels itselfs. it will be done on next state change.
    }

    void setIntervalLong(uint16_t newInterval) {
      intervalLong = newInterval;
    }

    void tick()    // time slot for the crossing manager
    {
      if (millis() - previousMillis > interval)
      {
        state++;
        state = state % 4;
        switch (state)
        {
          case 0:
            strip.setPixelColor(east0, trafficRed);
            strip.setPixelColor(east1, trafficRed);
            if (east2) strip.setPixelColor(east2, trafficRed);       // only if position for this traffic light was set
            if (east3) strip.setPixelColor(east3, trafficRed);
            strip.setPixelColor(north0, trafficGreen);
            strip.setPixelColor(north1, trafficGreen);
            if (north2) strip.setPixelColor(north2, trafficGreen);
            if (north3) strip.setPixelColor(north3, trafficGreen);
            interval = intervalLong;
            break;
          case 1:
          case 3:
            strip.setPixelColor(east0, trafficYellow);
            strip.setPixelColor(east1, trafficYellow);
            if (east2)strip.setPixelColor(east2, trafficYellow);
            if (east3)strip.setPixelColor(east3, trafficYellow);
            strip.setPixelColor(north0, trafficYellow);
            strip.setPixelColor(north1, trafficYellow);
            if (north2)strip.setPixelColor(north2, trafficYellow);
            if (north3)strip.setPixelColor(north3, trafficYellow);
            interval = intervalShort;
            break;
          default:
            strip.setPixelColor(east0, trafficGreen);
            strip.setPixelColor(east1, trafficGreen);
            if (east2)strip.setPixelColor(east2, trafficGreen);
            if (east3)strip.setPixelColor(east3, trafficGreen);
            strip.setPixelColor(north0, trafficRed);
            strip.setPixelColor(north1, trafficRed);
            if (north2)strip.setPixelColor(north2, trafficRed);
            if (north3)strip.setPixelColor(north3, trafficRed);
            interval = intervalLong;
        }
        strip.show();
        previousMillis = millis();
      }
    };
};


class Vehicle                                    // old/basic vehicle with a very simple pathfinding
{
  public:
    uint32_t color;                              // the color of this car
    bool debug = false;                          // debug messages for this car    Vehicle(uint32_t color): color(color) {}     // constructor
    int16_t newPosition = 4711;                  // number should be out of the range of the real grid.
    byte direction = random(north, west + 1);    // 0 north - 1 east - 2 south 3 west
    byte currentPosition = random(96, 159);      // startposition is anywhere in the middle of the screen, might interfere with some traffic lights, but I don't care...
    uint32_t previousMillis = 0;                 // timekeeping for intervals
    uint16_t wait = 500;                         // timekeeping standard/speed

    Vehicle(uint32_t color): color(color) {}     // constructor

    void setPosition(byte newPosition) {
      currentPosition = newPosition;
    }

    void setSpeed (uint16_t newSpeed) {
      wait = newSpeed;
    }

    void toggleDebug()
    {
      debug = !debug;
    }

    void tick()                                  // call this method in the loop (run)
    {
      if (millis() - previousMillis > wait)      // time to move
      {
        byte newDirection = direction;
        if (debug) Serial.print (color, HEX);
        for (byte i = 0; i < 4; i++)
        {
          if (pgm_read_byte(&street[currentPosition][i]) != 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][i])) == 0)
          {
            if (debug) {
              Serial.print(F(" cont dir="));
              Serial.print(i);
            }
            newPosition = pgm_read_byte(&street[currentPosition][i]);
            newDirection = i;
            break;
          }

          byte j = (i + 1) % 4;  // turn right/clockwise
          if (pgm_read_byte(&street[currentPosition][j]) != 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][j])) == 0)
          {
            if (debug) {
              Serial.print(F(" turnR dir="));
              Serial.print(j);
            }
            newPosition = pgm_read_byte(&street[currentPosition][j]);
            newDirection = j;
            break;
          }
          j = (i - 1) % 4;  // turn left/counter clockwise
          if (pgm_read_byte(&street[currentPosition][j]) != 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][j])) == 0)
          {
            if (debug) {
              Serial.print(F(" turnL dir="));
              Serial.print(j);
            }
            newPosition = pgm_read_byte(&street[currentPosition][j]);
            newDirection = j;
          }
        }

        if (newPosition != 0)          // do we have a new valid position?
        {
          if (debug) {
            Serial.print(F(" from ")); Serial.print(currentPosition);
            Serial.print(F(" to ")); Serial.println(newPosition);
          }
          if (newPosition != currentPosition) strip.setPixelColor(currentPosition, 0);   // delete if necessary
          strip.setPixelColor(newPosition, color);                                       // set new field
          strip.show();                                                                  // update strip to match
          currentPosition = newPosition;
          direction = newDirection;
        }
        else
        {
          if (debug)
          { Serial.print (F(" direction ")); Serial.print(direction);
            Serial.print(F(" blocked at ")); Serial.println(currentPosition);
          }
        }
        previousMillis = millis();
      }
    }
};


class Vehicle2000 : public Vehicle               // this vehicle have a sligthly improved movement decision
{
  public:
    using Vehicle::Vehicle;                      // baseclass has variables to initialize, therefore we need this explicit using

    void tick()
    {
      int16_t newPosition = -1;
      int8_t newDirection = -1;
      if (millis() - previousMillis > wait)
      {
        if (debug) Serial.print (color, HEX);    // use the color as identifier on the serial monitor for the car
        for (byte i = 0; i < 4; i++)
        {
          //Step 1: default keep direction
          if (direction == north && pgm_read_byte(&street[currentPosition][i]) > 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][i])) == 0)
          {
            newPosition = pgm_read_byte(&street[currentPosition][i]);
            break;
          }
          //Step 2: Any direction
          if (pgm_read_byte(&street[currentPosition][i]) > 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][i])) == 0 && random(0, 2)) // by random we do not use this direction
          {
            newPosition = pgm_read_byte(&street[currentPosition][i]);
            newDirection = i;
            break;
          }
          // Step3: if random() in step 2 avoided to find a proper direction, try again (counter clockwise)
          if (pgm_read_byte(&street[currentPosition][3 - i]) > 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][3 - i])) == 0)
          {
            newPosition = pgm_read_byte(&street[currentPosition][3 - i]);
            newDirection = 3 - i;
            break;
          }
        }

        if (newPosition > 0)           // we were able to get a new - free - position
        {
          if (debug) {
            Serial.print(F(" from ")); Serial.print(currentPosition);
            Serial.print(F(" to ")); Serial.println(newPosition);
          }
          if (newPosition != currentPosition)
          {
            strip.setPixelColor(currentPosition, 0);       // delete
            strip.setPixelColor(newPosition, color);       // set new field
            strip.show();                                  // update strip to match
          }
          currentPosition = newPosition;
          direction = newDirection;
        }
        else
        {
          if (debug)
          { Serial.print(F(" direction ")); Serial.print(direction);
            Serial.print(F(" blocked @")); Serial.println(currentPosition);
          }
        }
        previousMillis = millis();
      }
    }
};


class VehicleChaser : public Vehicle
{
  // this vehicle tries to get to a target point
  // if it can't use the shortest distance, it uses an alternative route to stay in motion
  public:
    using Vehicle::Vehicle;                      // wegen Initialisierungsliste der Basisklasse
    byte target = random(122, 133);
    byte order = 0;   // to be deleted tbd
    uint32_t previousMillisTarget;    // tbd
    byte mode = 0;                    // 0 idle; 1 chase

    void setTarget(byte newTarget, byte newOrder)
    {
      target = newTarget;
      order = newOrder;
      mode = 1;
      previousMillisTarget = millis();
    }

    void tick()
    {
      int16_t newPosition = -1;       // the new position
      int8_t newDirection = -1;       // the new direction
      uint16_t diffBest = 4711;       // best = shortest distance to target
      uint16_t diff = 4711;           // distance to target

      if (millis() - previousMillis > wait && mode != 2)
      {
        if (debug) {
          //Serial.println (strip.getPixelColor(currentPosition));
          Serial.print (color, HEX);    // I use the color as identifier for the car
          Serial.print (F(" target=")); Serial.print (target);
          Serial.print (F(" mode=")); Serial.print(mode);
        }
        for (byte i = 0; i < 4; i++)
        {
          //Step 1: default keep direction
          //if (direction == i && pgm_read_byte(&street[currentPosition][i]) > 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][i])) == 0) newPosition = pgm_read_byte(&street[currentPosition][i]);
          //Step 2: Any direction
          if (pgm_read_byte(&street[currentPosition][i]) > 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][i])) == 0)
            //if (pgm_read_byte(&street[currentPosition][i]) > 0)
          {
            diff = distance(pgm_read_byte(&street[currentPosition][i]), target);
            if (diff <= diffBest)
            {
              if (debug) {
                Serial.print(F(" dir"));
                Serial.print(i);
                Serial.print (F("="));
                Serial.print(diff);
              }
              newDirection = i;
              newPosition = pgm_read_byte(&street[currentPosition][i]);
              diffBest = diff;
            }
          }
        }

        if (newPosition > 0 && strip.getPixelColor(newPosition) == 0) // we were able to get a new - free - position
        {
          if (debug) {
            Serial.print(F(" from ")); Serial.print(currentPosition);
            Serial.print(F(" to ")); Serial.print(newPosition);
            Serial.print(F(" diff=")); Serial.print(diffBest);
          }
          if (newPosition != currentPosition)
          {
            strip.setPixelColor(currentPosition, 0);       // delete
            strip.setPixelColor(newPosition, color);       // set new field
            strip.show();                                  // update strip to match
          }
          currentPosition = newPosition;
          direction = newDirection;
          if (currentPosition == target && mode == 1)
          {
            if (debug) Serial.print(F(" | arrived "));
            mode = 2;                                      // MISSING: to be done: implement a really waiting time
          }
        }
        else
        {
          if (debug)
          { Serial.print(F(" direction ")); Serial.print(direction);
            Serial.print(F(" blocked @")); Serial.print(currentPosition);
          }
        }
        if (debug) Serial.println();
        previousMillis = millis();
      }
    }
};


class VehicleFiretruck : public Vehicle
{
  // this vehicle tries to get to a target point fast
  // idles slows
  // if it can't use the shortest distance, it waits till the tile gets unblocked
  public:
    using Vehicle::Vehicle;                      // wegen Initialisierungsliste der Basisklasse
    byte target = 147;
    byte order = 0;
    byte nextTarget;
    byte nextOrder;
    byte nextTarget1;
    byte nextOrder1;
    uint32_t previousMillisTarget;
    byte mode = 0;                    // 0 idle; 1 chase

    void setTarget(byte newTarget, byte newOrder)
    {
      if (mode == 0)                  // we were idling around, so we start a duty
      {
        target = newTarget;
        order = newOrder;
        mode = 1;
        previousMillisTarget = millis();
      }
      else
      {
        if (nextTarget == 0)
        {
          nextTarget = newTarget;         // if we are already on duty we store the incoming target as nextTarget
          nextOrder = newOrder;
        }
        else
        {
          nextTarget1 = newTarget;
          nextOrder1 = newOrder;
        }
      }
    }

    void tick()
    {
      int16_t newPosition = -1;       // the new position
      int8_t newDirection = -1;       // the new direction
      uint16_t diffBest = 4711;       // best = shortest distance to target
      uint16_t diff = 4711;           // distance to target

      if (millis() - previousMillis > (wait * (2 - mode)))   // in duty we speed up
      {
        if (debug) {
          //Serial.println (strip.getPixelColor(currentPosition));
          Serial.print (color, HEX);    // I use the color as identifier for the car
          Serial.print (F(" target=")); Serial.print (target);
          Serial.print (F(" mode=")); Serial.print(mode);
          if (nextTarget) {
            Serial.print (F(" next="));
            Serial.print(nextTarget);
          }
        }
        for (byte i = 0; i < 4; i++)
        {
          //Step 1: default keep direction
          //if (direction == i && pgm_read_byte(&street[currentPosition][i]) > 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][i])) == 0) newPosition = pgm_read_byte(&street[currentPosition][i]);
          //Step 2: Any direction
          //if (pgm_read_byte(&street[currentPosition][i]) > 0 && strip.getPixelColor(pgm_read_byte(&street[currentPosition][i])) == 0)
          if (pgm_read_byte(&street[currentPosition][i]) > 0)
          {
            diff = distance(pgm_read_byte(&street[currentPosition][i]), target);
            if (diff <= diffBest)    // if better keep current direction as best direction
            {
              if (debug) {
                Serial.print(F(" dir"));
                Serial.print(i);
                Serial.print (F("="));
                Serial.print(diff);
              }
              newDirection = i;
              newPosition = pgm_read_byte(&street[currentPosition][i]);
              diffBest = diff;
            }
          }
        }

        if (newPosition > 0 && strip.getPixelColor(newPosition) == 0) // we were able to get a new - free - position
        {
          if (debug) {
            Serial.print(F(" from ")); Serial.print(currentPosition);
            Serial.print(F(" to ")); Serial.print(newPosition);
            Serial.print(F(" diff=")); Serial.print(diffBest);
          }
          if (newPosition != currentPosition)
          {
            strip.setPixelColor(currentPosition, 0);       // delete
            strip.setPixelColor(newPosition, color);       // set new field
            strip.show();                                  // update strip to match
          }
          currentPosition = newPosition;
          direction = newDirection;

          if (currentPosition == target)                   // arrived at target
          {
            if (debug) Serial.print(F(" | arrived"));
            if (currentPosition == target) fireOut(target, order); // report success
            if (nextTarget)
            {
              target = nextTarget;
              order = nextOrder;
              nextOrder = nextOrder1;
              nextTarget = nextTarget1;
              nextOrder1 = 0;
              nextTarget1 = 0;
              previousMillisTarget = millis();
            }
            else
            {
              mode = 0;
              target = random(96, 159);  // just cruising around in the city center
            }
          }
        }
        else         // not free
        {
          if (debug)
          { Serial.print(F(" direction ")); Serial.print(direction);
            Serial.print(F(" blocked @")); Serial.print(currentPosition);
          }
        }

        if ((millis() - previousMillisTarget > 50000 && mode == 1)) // interrupt if we time is over
        {
          if (debug) Serial.print(F(" | timeout"));
          if (currentPosition == target) fireOut(target, order);    // report success
          if (nextTarget)
          {
            if (debug) Serial.print(F(" new target "));
            target = nextTarget;
            order = nextOrder;
            nextOrder = nextOrder1;
            nextTarget = nextTarget1;
            nextOrder1 = 0;
            nextTarget1 = 0;
            previousMillisTarget = millis();
          }
          else
          {
            mode = 0;
            target = random(96, 159);  // just cruising around in the city center
          }
        }

        if (debug) Serial.println();
        previousMillis = millis();
      }
    }
};


class Organisation 
{
  private:
    byte priority = 3;                        // the higher the less a target will be chosen, default 3

  public:
    uint32_t previousMillis = -300;           // timekeeping of object
    uint32_t previousMillisOrder = 0;         // timekeeping of the order
    VehicleChaser& obj;                       // a reference to the car/truck of the company
    byte action;                              // the action for the compay (company purpose)
    byte target = 0;                          // pixel where to go
    byte order = 0;                           // a order number (used as refernce when we vehicle reached target
    byte state = 0;                           // 0 ready // 1 Drive // 2 Arrival // 2 doing

    Organisation(VehicleChaser& obj, byte action) : obj(obj), action(action) {}     // the constructor takes a reference to the used vehicle

    void setPriority(byte newPriority)
    {
      priority = newPriority;
    }

    void tick()                               // simple logic for companies
    {
      if (millis() - previousMillis > 500)
      {
        if (state == 0)                        // find a new target
        {
          for (byte i = 0; i < noOfHouses; i++)
          {
            if (city.state[i] == action && random(0, priority) == 1)
            {
              target = pgm_read_byte(&houses[i][1]); // the street
              order = i;                             // the house-id (house array)
              //Serial.print(F("order for ")); Serial.println(order);
              break;
            }
          }
          if (target > 0)
          {
            obj.setTarget(target, order);
            state = 1;
            previousMillisOrder = millis();
          }
          else
          {
            obj.setTarget(255, 0);
          }
        }
        else if (state == 1)                                     // drive to target
        {
          if (obj.currentPosition == target)
          {
            // Serial.println(F("arrived @")); Serial.println(pgm_read_byte(&houses[order][0]));
            state = 2;
          }
          if (millis() - previousMillisOrder > 50000) state = 0; // check if order isn't to old ...
          if (city.state[order] != action) state = 0;            // check, if house has changed its state, then cancel order
        }
        else if (state == 2)
        {
          if (city.state[order] == action)                       // if house has still the right status, do your action
          {
            if (action == HOUSE)  // action on HOUSE has a different handling to get a fire
            {
              if (city.addDisaster (order))
              {
                Serial.print(F(" started fire @")); Serial.println(pgm_read_byte(&houses[order][0]));
              }
              else
              {
                Serial.print(F(" to late @")); Serial.println(pgm_read_byte(&houses[order][0]));
              }
            }
            else
            {
              city.state[order] = (action + 1) % 4;
              city.showHouse(order);
              //Serial.print(F("upgraded @")); Serial.println(pgm_read_byte(&houses[order][0]));
            }
          }
          state = 0;
        }
        previousMillis = millis();
      }
    }
};

// ********** create all the objects

// vehicles/cars
Vehicle2000   car0(0x804000);              // one separate car for testing
VehicleChaser carGardener(0x004000);       // a car for the gardener
VehicleChaser carBrickie(0xFFFFFF);        // a car for the brick layer
VehicleChaser carArsonist(0xFF0020);       // a car for the arsonsit

//VehicleFirefighter car1();               // a newer car for testing

Vehicle2000 cars[] 
{                             // a group of newer cars - to make some traffic in the city
  //  strip.Color(16,   0,   16),
  //  strip.Color(16,  10,   16),
  //  strip.Color(32,   0,   32),
  //  strip.Color(32,   20,  32),
  strip.Color(32,   30,  32),
  strip.Color(64,    0,  64),
  strip.Color(64,   16,  64),
  strip.Color(64,   32,  64),
  strip.Color(128,   0, 128),
  strip.Color(128,  16, 128),
  strip.Color(128,  32, 128),
  strip.Color(200,  16, 255),
  strip.Color(200,  32, 255),
  strip.Color(255,  0,  255),
  strip.Color(255,  16, 255),
  strip.Color(255,  32, 255)
};

Vehicle carsMore[] 
{              // a group of older cars
  0x20FF80,
  0x40FF80
};

VehicleFiretruck firetruck[] 
{              // a group of firetrucks
  0x0000FF,
  0x0000EE
};
const byte noOfFiretrucks = sizeof(firetruck)/sizeof(firetruck[0]);

Organisation gardenInc(carGardener, RUIN);      // the company for the gardener, using the green car, upgrades a RUIN to a PARK
Organisation brickieInc(carBrickie, PARK);      // the company of the brick layer, using a yellow car, upgrades a PARK to a HOUSE
Organisation spectre(carArsonist, HOUSE);       // SPECTRE organises the arsonist, HOUSE gets on FIRE

Crossing crossingCenter(86, 87, 167, 166, 101, 122, 138, 149);  // the main crossing in the center has 8 traffic lights (4x north/south, 4x east/west)
Crossing crossingNorth(39, 38, 21, 5);                          // the small crossings have only 4 traffic lights (2x north/south, 2x east/west)
Crossing crossingEast(94, 160, 130, 157);
Crossing crossingSouth(214, 215, 229, 245);
Crossing crossingWest(80, 174, 109, 114);

Button buttonPark(parkPin);
Button buttonRuin(ruinPin);
Button buttonFire(firePin);

// ********** some helper functions

// return a house of the desired state
// or the total number of houses if no house of this state available
byte findHouseByState(const byte state)
{
  byte actual = noOfHouses;
  for (byte i = 0; i < 5; i++)
  {
    actual = random(0, noOfHouses);
    if (city.state[actual] == state)
      return actual;
  }
  return actual;
}

// check the buttons and handle the action
void buttonTick()
{
  if (buttonPark.wasPressed())
  { byte actual = findHouseByState(RUIN);
    Serial.print(F("new PARK ")); Serial.println( actual);
    if (actual < noOfHouses)
      city.state[actual] = PARK;
    city.showHouse(actual);
  }
  if (buttonRuin.wasPressed())
  {
    byte actual = findHouseByState(HOUSE);
    Serial.print(F("new RUIN ")); Serial.println( actual);
    if (actual < noOfHouses)
      city.state[actual] = RUIN;
    city.showHouse(actual);
  }
  if (buttonFire.wasPressed())
  {
    byte actual = findHouseByState(HOUSE);
    Serial.print(F("new FIRE "));  Serial.println( actual);
    if (actual < noOfHouses) city.addDisaster(actual);
    city.showHouse(actual);
  }
}

// handle serial input
void serialTick()
{
  if (Serial.available())
  {
    char c = Serial.read();
    switch (c)
    {
      case 'a' : carArsonist.toggleDebug(); break;   // switch on/off debug messages for the arsonist's car
      case 'b' : carBrickie.toggleDebug(); break;
      case 'g' : carGardener.toggleDebug(); break;
      case 'F' : firetruck[0].toggleDebug(); break;
      case 'f' : firetruck[1].toggleDebug(); break;
      case 's' :
        Serial.print(F(" freeStack=")); Serial.println(freeStack()); break;
    }
  }
}


// ********** Arduino Standard: the setup:
void setup()
{
  strip.begin();             // INITIALIZE NeoPixel strip object (REQUIRED)
  strip.show();              // Turn OFF all pixels ASAP
  strip.setBrightness(8);    // Set BRIGHTNESS of NeoPixel (max = 255)
  Serial.begin(115200);
  Serial.println(F("\nNeopixel Traffic Simulator"));
  randomSeed(analogRead(0));

  buttonRuin.begin();
  buttonPark.begin();
  buttonFire.begin();

  showMap();                   // print the map

  // now to make the start somehow random, change some default values:
  //car0.setPosition(18);      // position of the car
  car0.setSpeed(235);
  //car0.debug = true;         // activate debug messages for this car
  //car1.setPosition(199);     // position of this car
  //car1.setSpeed(110);        // change speed of this car - 100 is very fast
  //car1.debug = true;         // activate debug messages for this car
  //carsMore[1].debug = true;  // dieses Auto krepiert manchmal auf Position 155

  carGardener.setSpeed(210);
  carBrickie.setSpeed(220);
  carArsonist.setSpeed(190);
  //carArsonist.debug = true;
  spectre.setPriority(2);      // 3 default, 1 no fires, 2 lot of fires, 6 some additional fires

  for (auto &i : cars)
    i.setSpeed((random(17, 28) * 10)); // give each car in this group an indiviudal speed

  for (auto &i : carsMore)
    i.setSpeed((random(20, 30) * 10));

  for (auto &i : firetruck)
  {
    i.setPosition(random(96, 159));
    i.setSpeed(110);
  }
  firetruck[0].debug = true;         // activate debug messages for this car

  // make different intervals and start states for the crossings
  crossingCenter.setIntervalLong(3090);
  crossingNorth.setIntervalLong(4010);
  crossingEast.setIntervalLong(4520);
  crossingSouth.setIntervalLong(5530);
  crossingWest.setIntervalLong(6040);
  crossingCenter.setState(2);
  crossingNorth.setState(1);
  crossingEast.setState(2);
  crossingSouth.setState(3);

  for (byte i = 0; i < noOfHouses; i++)
    city.state[i] = random(0, 3);  // oder fix HOUSE
  city.showAllHouses();
}

// ********** Arduino Standard: the main loop
void loop()
{
  crossingCenter.tick();     // tick the crossing so it can update the state if necessary
  crossingNorth.tick();
  crossingEast.tick();
  crossingSouth.tick();
  crossingWest.tick();
  city.tick();               // The city handles the houses (and the desasters)

  car0.tick();               // just tick the cars - if needed they will move
  //car1.tick();
  for (auto &i : cars)       // there are lot of cars in this group/array of objects
    i.tick();
  for (auto &i : carsMore)
    i.tick();
  for (auto &i : firetruck)
    i.tick();

  carGardener.tick();
  carBrickie.tick();
  carArsonist.tick();

  brickieInc.tick();
  gardenInc.tick();
  spectre.tick();

  serialTick();
  buttonTick();
}


// some functions - still to be migrated to OOP
void firedepartmentAlert(byte street, byte order)
{
  // The firedepartment distributes the incoming alerts to the firetrucks.
  static byte actual;
  actual++;
  actual = actual % noOfFiretrucks;
  firetruck[actual].setTarget(street, order);
}

void fireOut(byte street, byte order)
{
  city.receivedFireOut(street, order);
}


#ifdef __arm__
// should use uinstd.h to define sbrk but Due causes a conflict
extern "C" char* sbrk(int incr);
#else  // __ARM__
extern char *__brkval;
#endif  // __arm__

int freeStack() {
  char top;
#ifdef __arm__
  return &top - reinterpret_cast<char*>(sbrk(0));
#elif defined(CORE_TEENSY) || (ARDUINO > 103 && ARDUINO != 151)
  return &top - __brkval;
#else  // __arm__
  return __brkval ? &top - __brkval : &top - __malloc_heap_start;
#endif  // __arm__
}
